#ifdef HAVE_CONFIG_H
#include "mysqludf_config.h"
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif

#include "duration.h"

#ifndef local
#define local static inline
#endif

#define ily(y) ((y) % 4 == 0 && ((y) % 100 != 0 || (y) % 400 == 0))

#define adj_gen(inc, val, nxt) \
	if (val < 0 || val > inc) { \
		nxt += val / inc ? val / inc : val > 0 ? 1 : -1; \
		val = inc - (val > 0 ? 1 : -1 * (val % inc)); \
	}

#define adj_mon(m, y) \
	while (m >= 12) { \
		m -= 12; \
		y += 1; \
	} \
	while (m < 0) { \
		m += 12; \
		y -= 1; \
	}

local int dim(int m, int y)
{
	static const char days_in_month[2][12] = {
		{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
		{31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
	};
	
	adj_mon(m, y);
	
	return days_in_month[ily(y)][m];
}

local struct tm *adj(struct tm *t)
{
	adj_gen(60, t->tm_sec, t->tm_min);
	adj_gen(60, t->tm_min, t->tm_hour);
	adj_gen(24, t->tm_hour, t->tm_mday);
	
	while (t->tm_mday > dim(t->tm_mon, t->tm_year)) {
		t->tm_mday -= dim(t->tm_mon, t->tm_year);
		t->tm_mon  += 1;
		adj_mon(t->tm_mon, t->tm_year);
	}
	while (t->tm_mday < 1) {
		t->tm_mon  -= 1;
		adj_mon(t->tm_mon, t->tm_year);
		t->tm_mday += dim(t->tm_mon, t->tm_year);
	}
}

local sll sec(struct tm *t1, struct tm *t2)
{
	return t1->tm_hour * 60 * 60 + t1->tm_min * 60 + t1->tm_sec + 24 * 60 * 60 - t2->tm_hour * 60 * 60 - t2->tm_min * 60 - t2->tm_sec;
}

local sll day(struct tm *t1, struct tm *t2)
{
	int y, m, d;
	
	if (t1->tm_year == t2->tm_year && t1->tm_mon == t2->tm_mon) {
		d = t1->tm_mday - t2->tm_mday;
	} else {
		d = t1->tm_mday + dim(t2->tm_mon, t2->tm_year + 1900) - t2->tm_mday;
		for (y = t2->tm_year; y <= t1->tm_year; ++y) {
			for (m = (y == t2->tm_year) ? t2->tm_mon + 1 : 0; (y == t2->tm_year && y != t1->tm_year) ? m < 12 : m < t1->tm_mon; ++m) {
				d += dim(m, y + 1900);
			}
		}
	}
	
	if (d && sec(t1, t2) < 24 * 60 * 60) {
		d -= 1;
	}
	
	return d;
}

local sll mon(struct tm *t1, struct tm *t2)
{
	if (t1->tm_year == t2->tm_year) {
		return t1->tm_mon - t2->tm_mon;
	} else {
		return (t1->tm_year - t2->tm_year - 1) * 12 + t1->tm_mon + 12 - t2->tm_mon;
	}
}

struct tm *sll2tm(struct tm *t, sll date)
{
#define _calc(_tvar, _date, _f) \
	_tvar = _date % _f; \
	_date = (_date - _tvar) / _f;
	
	_calc(t->tm_sec, date, 100);
	_calc(t->tm_min, date, 100);
	_calc(t->tm_hour, date, 100);
	_calc(t->tm_mday, date, 100);
	_calc(t->tm_mon, date, 100);
	_calc(t->tm_year, date, 10000);
	
	t->tm_mon  -= 1;
	t->tm_year -= 1900;
	
	return adj(t);
}

sll duration(struct tm *t1, struct tm *t2, duration_precision_t w)
{
	switch (w) {
		case DURATION_PRECISION_Y:
			return mon(t1, t2) / SLL(12);
		case DURATION_PRECISION_M:
			return mon(t1, t2);
		case DURATION_PRECISION_D:
			return day(t1, t2);
		case DURATION_PRECISION_H:
			return day(t1, t2) * SLL(24) + sec(t1, t2) / SLL(3600);
		case DURATION_PRECISION_I:
			return day(t1, t2) * SLL(24) * SLL(60) + sec(t1, t2) / SLL(60);
		case DURATION_PRECISION_S:
			return day(t1, t2) * SLL(24) * SLL(60) * SLL(60) + sec(t1, t2);
	}
}

sll duration_y(struct tm *t1, struct tm *t2)
{
	return mon(t1, t2) / SLL(12);
}

sll duration_m(struct tm *t1, struct tm *t2)
{
	return mon(t1, t2);
}

sll duration_d(struct tm *t1, struct tm *t2)
{
	return day(t1, t2);
}

sll duration_h(struct tm *t1, struct tm *t2)
{
	return day(t1, t2) * SLL(24) + sec(t1, t2) / SLL(3600);
}

sll duration_i(struct tm *t1, struct tm *t2)
{
	return day(t1, t2) * SLL(24) * SLL(60) + sec(t1, t2) / SLL(60);
}

sll duration_s(struct tm *t1, struct tm *t2)
{
	return day(t1, t2) * SLL(24) * SLL(60) * SLL(60) + sec(t1, t2);
}
